from contextlib import asynccontextmanager
from functools import partial
from pathlib import Path
from typing import Annotated, Any

from anyio import sleep_forever
from anyio.abc import TaskStatus
from fastapi import Cookie, Depends, FastAPI, WebSocket, WebSocketDisconnect, status
from fastapi_users import BaseUserManager, models
from pycrdt import Channel
from wiredb import Room, RoomManager, connect

from .db import create_db_and_tables
from .schemas import UserCreate, UserRead, UserUpdate
from .users import auth_backend, fastapi_users, get_user_manager, get_jwt_strategy


class StoredRoom(Room):
    def __init__(self, directory: str, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._directory = directory

    async def run(self, *args: Any, **kwargs: Any):
        await self.task_group.start(self.connect_to_file)
        await super().run(*args, **kwargs)

    async def connect_to_file(self, *, task_status: TaskStatus[None]) -> None:
        async with connect("file", doc=self.doc, path=f"{Path(self._directory) / self.id.lstrip('/')}.y"):
            task_status.started()
            await sleep_forever()


class CocatApp:
    def __init__(self, update_dir: str, db_path: str = "./test.db") -> None:

        @asynccontextmanager
        async def lifespan(app: FastAPI):
            async with RoomManager(partial(StoredRoom, update_dir)) as self.room_manager:
                await create_db_and_tables(db_path)
                yield

        self.app = app = FastAPI(lifespan=lifespan)

        current_superuser = fastapi_users.current_user(active=True, superuser=True)

        app.include_router(
            fastapi_users.get_auth_router(auth_backend), prefix="/auth/jwt", tags=["auth"]
        )
        app.include_router(
            fastapi_users.get_register_router(UserRead, UserCreate),
            prefix="/auth",
            tags=["auth"],
            dependencies=[Depends(current_superuser)]
        )
        app.include_router(
            fastapi_users.get_reset_password_router(),
            prefix="/auth",
            tags=["auth"],
        )
        app.include_router(
            fastapi_users.get_verify_router(UserRead),
            prefix="/auth",
            tags=["auth"],
        )
        app.include_router(
            fastapi_users.get_users_router(UserRead, UserUpdate),
            prefix="/users",
            tags=["users"],
        )

        @app.websocket("/room/{id}")
        async def connect_room(
            id: str,
            websocket = Depends(websocket_auth),
        ):
            if websocket is None:
                return

            await websocket.accept()
            ywebsocket = YWebSocket(websocket, id)
            room = await self.room_manager.get_room(ywebsocket.path)
            await room.serve(ywebsocket)


async def websocket_auth(
    websocket: WebSocket,
    fastapiusersauth: Annotated[str | None, Cookie()] = None,
    user_manager: BaseUserManager[models.UP, models.ID] = Depends(get_user_manager),
) -> WebSocket | None:
    accept_websocket = False
    if fastapiusersauth is not None:
        user = await get_jwt_strategy().read_token(fastapiusersauth, user_manager)  # type: ignore[func-returns-value,arg-type]
        if user:
            accept_websocket = True
    if accept_websocket:
        return websocket

    await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
    return None


class YWebSocket(Channel):
    def __init__(self, websocket: WebSocket, path: str) -> None:
        self._websocket = websocket
        self._path = path

    @property
    def path(self) -> str:
        return self._path

    async def __anext__(self):
        try:
            return await self._websocket.receive_bytes()
        except WebSocketDisconnect:
            raise StopAsyncIteration()

    async def send(self, message: bytes) -> None:
        await self._websocket.send_bytes(message)

    async def recv(self) -> bytes:
        return await self._websocket.receive_bytes()



from .db import DB as DB
from .catalogue import Catalogue as Catalogue
from .event import Event as Event
from .models import CatalogueModel as CatalogueModel
from .models import EventModel as EventModel
from .api import create_catalogue as create_catalogue
from .api import create_event as create_event
from .api import load_catalogue as load_catalogue
from .api import load_event as load_event
from .api import log_in as log_in
from .api import log_out as log_out
from .api import save_catalogue as save_catalogue
from .api import save_event as save_event
from .api import set_config as set_config



from collections.abc import AsyncGenerator

from anyio import Lock
from fastapi import Depends
from fastapi_users.db import SQLAlchemyBaseUserTableUUID, SQLAlchemyUserDatabase
from fastapi_users_db_sqlalchemy.access_token import (
    SQLAlchemyAccessTokenDatabase,
    SQLAlchemyBaseAccessTokenTableUUID,
)
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import DeclarativeBase

async_session_maker: async_sessionmaker[AsyncSession] | None = None
lock = Lock()


class Base(DeclarativeBase):
    pass


class User(SQLAlchemyBaseUserTableUUID, Base):
    pass


class AccessToken(SQLAlchemyBaseAccessTokenTableUUID, Base):  
    pass


async def create_db_and_tables(path: str):
    global async_session_maker

    async with lock:
        if async_session_maker is not None:
            return

        database_url = f"sqlite+aiosqlite:///{path}"
        engine = create_async_engine(database_url)
        async_session_maker = async_sessionmaker(engine, expire_on_commit=False)

        async with engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)


async def get_async_session() -> AsyncGenerator[AsyncSession, None]:
    assert async_session_maker is not None
    async with async_session_maker() as session:
        yield session


async def get_user_db(session: AsyncSession = Depends(get_async_session)):
    yield SQLAlchemyUserDatabase(session, User)


async def get_access_token_db(
    session: AsyncSession = Depends(get_async_session),
):
    yield SQLAlchemyAccessTokenDatabase(session, AccessToken)



from datetime import datetime
from typing import Any
from uuid import UUID, uuid4

from pydantic import BaseModel, Field


class EventModel(BaseModel):
    """
    The [pydantic](https://docs.pydantic.dev) model for event validation.
    """
    uuid: UUID = Field(default_factory=lambda: uuid4())
    start: datetime
    stop: datetime
    author: str
    tags: list[str] = Field(default_factory=list)
    products: list[str] = Field(default_factory=list)
    rating: int | None = None
    attributes: dict[str, Any] = Field(default_factory=dict)


class CatalogueModel(BaseModel):
    """
    The [pydantic](https://docs.pydantic.dev) model for catalogue validation.
    """
    uuid: UUID = Field(default_factory=lambda: uuid4())
    name: str
    author: str
    tags: list[str] = Field(default_factory=list)
    events: list[str] = Field(default_factory=list)
    attributes: dict[str, Any] = Field(default_factory=dict)



from datetime import datetime
from collections.abc import Iterable
from typing import Any
from uuid import UUID

import httpx
from anyio import Lock
from cocat import DB, Catalogue, Event
from pycrdt import Doc
from wiredb import connect


class Session:
    def __init__(self, host: str = "http://localhost", port: int = 8000, file_path: str = "updates.y", room_id: str = "room0"):
        self.catalogues: dict[str, Catalogue] = {}
        self.events: dict[str, Event] = {}
        self.host = host
        self.port = port
        self.cookies = httpx.Cookies()
        self.file_path = file_path
        self.room_id = room_id
        self.lock = Lock()

    async def connect(self, doc: Doc) -> None:
        async with self.lock:
            async with connect("websocket", id=f"room/{self.room_id}", doc=doc, host=self.host, port=self.port, cookies=self.cookies) as self.client:
                async with connect("file", doc=doc, path=self.file_path) as self.file:
                    pass

    def create_catalogue(
        self,
        name: str,
        author: str,
        uuid: UUID | str | bytes | bytearray | None = None,
        tags: list[str] | None = None,
        attributes: dict[str, Any] | None = None,
        events: Iterable[Event] | Event | None = None,
    ):
        db = DB()
        catalogue = db.create_catalogue(
            name=name,
            author=author,
            uuid=uuid,
            tags=tags,
            attributes=attributes,
            events=events,
        )
        self.catalogues[str(catalogue.uuid)] = catalogue
        self.catalogues[catalogue.name] = catalogue
        return catalogue

    def create_event(
        self,
        start: datetime | int | float | str,
        stop: datetime | int | float | str,
        author: str,
        uuid: UUID | str | bytes | bytearray | None = None,
        tags: list[str] | None = None,
        products: list[str] | None = None,
        rating: int | None = None,
        attributes: dict[str, Any] | None = None,
    ):
        db = DB()
        event = db.create_event(
            start=start,
            stop=stop,
            author=author,
            uuid=uuid,
            tags=tags,
            products=products,
            rating=rating,
            attributes=attributes,
        )
        self.events[str(event.uuid)] = event
        return event

    def get_local_catalogue(self, uuid_or_name: str) -> Catalogue:
        return self.catalogues[uuid_or_name]

    async def get_remote_catalogue(self, uuid_or_name: str) -> Catalogue:
        db = DB()
        await self.connect(db.doc)
        catalogue = db.get_catalogue(uuid_or_name)
        self.catalogues[str(catalogue.uuid)] = catalogue
        self.catalogues[str(catalogue.name)] = catalogue
        return catalogue

    def get_local_event(self, uuid: str) -> Event:
        return self.events[uuid]

    async def get_remote_event(self, uuid: str) -> Event:
        db = DB()
        await self.connect(db.doc)
        event = db.get_event(uuid)
        self.events[str(event.uuid)] = event
        return event


SESSION = Session()


def set_config(*, host: str | None = None, port: int | None = None, file_path: str | None = None, room_id: str | None = None) -> None:
    """
    Sets the configuration of the current session.

    Args:
        host: The host name of the database web server.
        port: The port number of the database web server.
        file_path: The path to the file where updates will be stored.
        room_id: The ID of the room to connect to.
    """
    if host is not None:
        SESSION.host = host
    if port is not None:
        SESSION.port = port
    if file_path is not None:
        SESSION.file_path = file_path
    if room_id is not None:
        SESSION.room_id = room_id


def log_in(username: str, password: str) -> None:
    """
    Log into the server.

    Args:
        username: The username to use to log in.
        password: The password to use to log in.
    """
    data = {"username": username, "password": password}
    response = httpx.post(f"{SESSION.host}:{SESSION.port}/auth/jwt/login", data=data)
    cookie = response.cookies.get("fastapiusersauth")
    assert cookie is not None
    SESSION.cookies.set("fastapiusersauth", cookie)


def log_out() -> None:
    """
    Log out of the server.
    """
    httpx.post(f"{SESSION.host}:{SESSION.port}/auth/jwt/logout", cookies=SESSION.cookies)
    SESSION.cookies = httpx.Cookies()


def create_catalogue(
    *,
    name: str,
    author: str,
    uuid: UUID | str | bytes | bytearray | None = None,
    tags: list[str] | None = None,
    attributes: dict[str, Any] | None = None,
    events: Iterable[Event] | Event | None = None,
) -> Catalogue:
    """
    Creates a catalogue in the database.

    Args:
        name: The name of the catalogue.
        author: The author of the catalogue.
        uuid: The optional UUID of the catalogue.
        tags: The optional tags of the catalogue.
        attributes: The optional attributes of the catalogue.
        events: The initial event(s) in the catalogue.

    Returns:
        The created [Catalogue][cocat.Catalogue].
    """
    return SESSION.create_catalogue(
        name=name,
        author=author,
        uuid=uuid,
        tags=tags,
        attributes=attributes,
        events=events,
    )


def create_event(
    *,
    start: datetime | int | float | str,
    stop: datetime | int | float | str,
    author: str,
    uuid: UUID | str | bytes | bytearray | None = None,
    tags: list[str] | None = None,
    products: list[str] | None = None,
    rating: int | None = None,
    attributes: dict[str, Any] | None = None,
) -> Event:
    """
    Creates an event in the database.

    Args:
        start: The start date of the event.
        stop: The stop date of the event.
        author: The author of the event.
        uuid: The optional UUID of the event.
        tags: The optional tags of the event.
        products: The optional products of the event.
        rating: The optional rating of the event.
        attributes: The optional attributes of the catalogue.

    Returns:
        The created [Event][cocat.Event].
    """
    return SESSION.create_event(
        start=start,
        stop=stop,
        author=author,
        uuid=uuid,
        tags=tags,
        products=products,
        rating=rating,
        attributes=attributes,
    )


async def load_catalogue(uuid_or_name: UUID | str) -> Catalogue:
    """
    Loads a catalogue from the server.

    Args:
        uuid_or_name: The UUID or the name of the catalogue to load.

    Returns:
        The loaded catalogue.
    """
    return await SESSION.get_remote_catalogue(str(uuid_or_name))


async def save_catalogue(catalogue: Catalogue | UUID | str) -> None:
    """
    Saves a catalogue in the server.

    Args:
        catalogue: The catalogue to save (or its UUID).
    """
    if isinstance(catalogue, Catalogue):
        uuid_or_name = str(catalogue.uuid)
    else:
        uuid_or_name = str(catalogue)
    catalogue = SESSION.get_local_catalogue(uuid_or_name)
    await SESSION.connect(catalogue.db.doc)


async def load_event(uuid: UUID | str) -> Event:
    """
    Loads an event from the server.

    Args:
        uuid: The UUID of the event to load.

    Returns:
        The loaded event.
    """
    return await SESSION.get_remote_event(str(uuid))


async def save_event(event: Event | UUID | str) -> None:
    """
    Saves an event in the server.

    Args:
        event: The event to save (or its UUID).
    """
    if isinstance(event, Event):
        uuid = str(event.uuid)
    else:
        uuid = str(event)
    event = SESSION.get_local_event(uuid)
    await SESSION.connect(event.db.doc)



import sys
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from json import dumps
from typing import Any, TYPE_CHECKING, cast

from pycrdt import Map
from simpleeval import SimpleEval  # type: ignore[import-untyped]

from .base import Mixin
from .event import Event
from .models import CatalogueModel

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: nocover
    from typing_extensions import Self

if TYPE_CHECKING:
    from .db import DB


@dataclass(eq=False)
class Catalogue(Mixin):
    _uuid: str
    _map: Map
    _db: "DB"
    _condition: str | None = None

    def _check_deleted(self):
        if self._uuid not in self._db._catalogue_maps:
            raise RuntimeError("Catalogue has been deleted")

    def __eq__(self, other: Any) -> bool:
        self._check_deleted()
        if not isinstance(other, Catalogue):
            return NotImplemented

        return self._uuid == other._uuid

    def __repr__(self) -> str:
        return dumps(self.to_dict())

    def __hash__(self) -> int:
        return hash(self._uuid)

    def __contains__(self, event: Event) -> bool:
        return str(event.uuid) in self._map["events"]

    def _get(self, name: str) -> Any:
        self._check_deleted()
        value = self._map[name]
        model = CatalogueModel.__pydantic_validator__.validate_assignment(CatalogueModel.model_construct(), name, value)
        return getattr(model, name)

    def _set(self, name: str, value: Any) -> None:
        with self._db.transaction():
            self._check_deleted()
            model = CatalogueModel.__pydantic_validator__.validate_assignment(CatalogueModel.model_construct(), name, value)
            val = getattr(model, name)
            self._map[name] = val

    def _on_change(self, name: str, callback: Callable[[Any], None]) -> None:
        self._check_deleted()
        self._db._catalogue_change_callbacks[self._uuid][name].append(partial(self._callback, callback))

    def _on_add(self, field: str, callback: Callable[[Any], None]) -> None:
        self._check_deleted()
        self._db._catalogue_change_callbacks[self._uuid][f"add_{field}"].append(partial(self._callback, callback))

    def _on_remove(self, field: str, callback: Callable[[list[str]], None]) -> None:
        self._check_deleted()
        self._db._catalogue_change_callbacks[self._uuid][f"remove_{field}"].append(partial(self._callback, callback))

    @classmethod
    def new(cls, model: CatalogueModel, db: "DB") -> Self:
        uuid = str(model.uuid)
        map = Map(dict(
            uuid=uuid,
            name=model.name,
            author=model.author,
            tags=Map({val: True for val in model.tags}),
            events=Map({val: True for val in model.events}),
            attributes=Map(model.attributes),
        ))
        self = cls(uuid, map, db)
        db._catalogues[uuid] = self
        return self

    @classmethod
    def from_map(cls, map: Map, db: "DB") -> Self:
        uuid = map["uuid"]
        self = cls(uuid, map, db)
        db._catalogues[uuid] = self
        return self

    @classmethod
    def from_uuid(cls, uuid: str, db: "DB") -> Self:
        map = db._catalogue_maps[uuid]
        self = cls(uuid, map, db)
        db._catalogues[uuid] = self
        return self

    def to_dict(self) -> dict[str, Any]:
        """
        Returns:
            The catalogue as a dictionary.
        """
        self._check_deleted()
        dct = self._map.to_py()
        assert dct is not None
        dct["tags"] = list(dct["tags"].keys())
        dct["events"] = list(dct["events"].keys())
        dct["attributes"] = dict(sorted(dct["attributes"].items()))
        return dict(sorted(dct.items()))

    def on_change_name(self, callback: Callable[[str], None]) -> None:
        """
        Registers a callback to be called when the catalogue name changes.

        Args:
            callback: The callback to call with the new name.
        """
        self._on_change("name", callback)

    def on_change_author(self, callback: Callable[[Any], None]) -> None:
        """
        Registers a callback to be called when the catalogue author changes.

        Args:
            callback: The callback to call with the new author.
        """
        self._on_change("author", callback)

    def on_delete(self, callback: Callable[[], None]) -> None:
        """
        Registers a callback to be called when the catalogue is removed from the database.

        Args:
            callback: The callback to call.
        """
        self._check_deleted()
        self._db._catalogue_delete_callbacks[self._uuid].append(partial(self._callback, callback))

    def delete(self) -> None:
        """
        Removes the catalogue from the database.
        """
        with self._db.transaction():
            self._check_deleted()
            del self._db._catalogue_maps[self._uuid]

    def on_add_events(self, callback: Callable[[list[Event]], None]) -> None:
        """
        Registers a callback to be called when events are added to the catalogue.

        Args:
            callback: The callback to call with a list of added events.
        """
        self._on_add("events", callback)

    def on_remove_events(self, callback: Callable[[list[str]], None]) -> None:
        """
        Registers a callback to be called when events are removed from the catalogue.

        Args:
            callback: The callback to call with a list of removed event UUIDs.
        """
        self._on_remove("events", callback)

    def add_events(self, events: Iterable[Event] | Event) -> None:
        """
        Add event(s) to the catalogue.

        Args:
            events: The events to add to the catalogue.
        """
        with self._db.transaction():
            event_list = [events] if isinstance(events, Event) else events
            self._check_deleted()
            map = cast(Map, self._map["events"])
            for event in event_list:
                map[event._uuid] = True

    def set_dynamic_filter(
        self,
        condition: str | None = None,
    ) -> None:
        """
        Sets a condition that will be evaluated when accessing `catalogue.dynamic_events`.
        The condition is an expression using the event attributes and/or references to other catalogues, for instance:
        ```py
        catalogue.set_dynamic_filter(f"event.start > datetime(2025, 1, 1) and event.stop <= datetime(2026, 1, 1) and event in catalogue('my_catalogue_name_or_uuid')")
        ```

        Args:
            condition: The condition an event needs to match to be part of the catalogue.
        """
        self._condition = condition

    def remove_events(self, events: Iterable[Event] | Event) -> None:
        """
        Removes event(s) from the catalogue.

        Args:
            events: The events to remove from the catalogue.
        """
        with self._db.transaction():
            event_list = [events] if isinstance(events, Event) else events
            self._check_deleted()
            map = cast(Map, self._map["events"])
            for event in event_list:
                del map[event._uuid]

    @property
    def name(self) -> str:
        """
        Returns:
            The name of the catalogue.
        """
        return self._get("name")

    @name.setter
    def name(self, value: str) -> None:
        """
        Args:
            value: The name of the catalogue to set.
        """
        self._set("name", value)

    @property
    def dynamic_events(self) -> set[Event]:
        """
        Returns:
            The dynamic events in the catalogue, as defined by `catalogue.set_dynamic_filter(condition)`.
        """
        if not self._condition:
            return set()

        s = SimpleEval()
        s.functions = {"datetime": datetime, "catalogue": self._db.get_catalogue}
        events = set()
        for event in self._db.events:
            s.names = {"event": event}
            if s.eval(self._condition):
                events.add(event)

        return events

    @property
    def all_events(self) -> set[Event]:
        """
        Returns:
            The (static and dynamic) events in the catalogue.
        """
        return self.events.union(self.dynamic_events)

    @property
    def events(self) -> set[Event]:
        """
        Returns:
            The (static) events in the catalogue.
        """
        self._check_deleted()
        event_uuids = cast(Map, self._map["events"])
        return {Event.from_map(self._db._event_maps[uuid], self._db) for uuid in event_uuids.keys()}

    @events.setter
    def events(self, value: set[Event]) -> None:
        """
        Args:
            value: The events to set in the catalogue.
        """
        with self._db.transaction():
            self._check_deleted()
            events = cast(Map, self._map["events"])
            events.clear()
            for event in value:
                self.add_events(event)



from __future__ import annotations

import json
from collections import defaultdict
from collections.abc import Callable, Iterable
from datetime import datetime
from functools import partial
from typing import Any
from uuid import UUID

from pycrdt import (
    ArrayEvent,
    Doc,
    Map,
    MapEvent,
    Transaction,
    TransactionEvent,
    YMessageType,
    create_sync_message,
    create_update_message,
    handle_sync_message,
)

from .catalogue import Catalogue
from .event import Event
from .models import CatalogueModel, EventModel


class DB:
    """
    A database which holds events and catalogues.
    """
    def __init__(self, doc: Doc | None = None) -> None:
        """
        Creates a database.

        Args:
            doc: An optional [Doc](https://y-crdt.github.io/pycrdt/api_reference/#pycrdt.Doc).
        """
        self._doc: Doc = Doc() if doc is None else doc
        self._catalogue_maps = self._doc.get("catalogues", type=Map)
        self._event_maps = self._doc.get("events", type=Map)
        self._synced: list[DB] = []
        self._catalogue_maps.observe_deep(self._catalogues_changed)
        self._catalogue_delete_callbacks: dict[str, list[Callable[[Any], None]]] = defaultdict(list)
        self._catalogue_create_callbacks: list[Callable[[Any, Any], None]] = []
        self._catalogue_change_callbacks: dict[str, dict[str, list[Callable[[Any, Any], None]]]] = defaultdict(lambda: defaultdict(list))
        self._catalogues: dict[str, Catalogue] = {}
        self._event_maps.observe_deep(self._events_changed)
        self._event_delete_callbacks: dict[str, list[Callable[[Any], None]]] = defaultdict(list)
        self._event_create_callbacks: list[Callable[[Any, Any], None]] = []
        self._event_change_callbacks: dict[str, dict[str, list[Callable[[Any, Any], None]]]] = defaultdict(lambda: defaultdict(list))
        self._events: dict[str, Event] = {}

    def _callback(self, callback: Callable[..., None], origin: "DB" | None, *args: Any) -> None:
        if origin is not self:
            callback(*args)

    def transaction(self) -> Transaction:
        return self._doc.transaction(self)

    @classmethod
    def from_json(cls, data: str, doc: Doc | None = None) -> "DB":
        """
        Creates a database from a JSON string.

        Args:
            data: The JSON string.
            doc: An optional [Doc](https://y-crdt.github.io/pycrdt/api_reference/#pycrdt.Doc).

        Returns:
            The created database.
        """
        db = DB(doc=doc)
        with db.transaction():
            db_dict = json.loads(data)
            for item in db_dict["events"]:
                db.create_event(**item)
            for item in db_dict["catalogues"]:
                events = [db.get_event(uuid) for uuid in item.pop("events", [])]
                db.create_catalogue(events=events, **item)
            return db

    @property
    def doc(self) -> Doc:
        """
        Returns:
            The database [Doc](https://y-crdt.github.io/pycrdt/api_reference/#pycrdt.Doc).
        """
        return self._doc

    def _catalogues_changed(self, events: list[ArrayEvent | MapEvent], transaction: Transaction) -> None:
        for event in events:
            path = event.path  # type: ignore[union-attr]
            if len(path) == 0:
                # catalogue created or deleted
                assert isinstance(event, MapEvent)
                keys = event.keys  # type: ignore[attr-defined]
                for uuid in keys:
                    action = keys[uuid]["action"]
                    if action == "delete":
                        for delete_callback in self._catalogue_delete_callbacks[uuid]:
                            delete_callback(transaction.origin)
                        if uuid in self._catalogues:
                            del self._catalogues[uuid]
                        del self._catalogue_delete_callbacks[uuid]
                        self._catalogue_change_callbacks
                        del self._catalogue_change_callbacks[uuid]
                    elif action == "add":
                        for create_callback in self._catalogue_create_callbacks:
                            create_callback(transaction.origin, self.get_catalogue(uuid))
            elif len(path) == 1:
                # property of catalogue changed (not events)
                assert isinstance(event, MapEvent)
                uuid = path[0]
                changed_keys = event.keys  # type: ignore[attr-defined]
                for key in changed_keys:
                    if key in self._catalogue_change_callbacks[uuid]:
                        callbacks = self._catalogue_change_callbacks[uuid][key]
                        for callback in callbacks:
                            value = changed_keys[key]["newValue"]
                            model = CatalogueModel.__pydantic_validator__.validate_assignment(CatalogueModel.model_construct(), key, value)
                            callback(transaction.origin, getattr(model, key))
            elif len(path) == 2:
                if path[1] == "events":
                    # catalogue events changed
                    assert isinstance(event, MapEvent)
                    uuid = path[0]
                    if (
                        "add_events" in self._catalogue_change_callbacks[uuid] or
                        "remove_events" in self._catalogue_change_callbacks[uuid]
                    ):
                        added_uuids = []
                        removed_uuids = []
                        keys = event.keys  # type: ignore[attr-defined]
                        for key, val in keys.items():
                            if val["action"] == "delete":
                                removed_uuids.append(key)
                            else:
                                added_uuids.append(key)
                        if removed_uuids:
                            callbacks = self._catalogue_change_callbacks[uuid]["remove_events"]
                            for callback in callbacks:
                                callback(transaction.origin, set(removed_uuids))
                        if added_uuids:
                            result = {Event.from_map(self._event_maps[added_uuid], self) for added_uuid in added_uuids}
                            callbacks = self._catalogue_change_callbacks[uuid]["add_events"]
                            for callback in callbacks:
                                callback(transaction.origin, result)
                else:
                    assert isinstance(event, MapEvent)
                    uuid, name = path
                    added = {}
                    removed = set()
                    keys = event.keys  # type: ignore[attr-defined]
                    for key, val in keys.items():
                        if val["action"] == "delete":
                            removed.add(key)
                        elif val["action"] == "add":
                            added[key] = val["newValue"]
                        elif val["action"] == "update":
                            added[key] = val["newValue"]
                    if removed:
                        callbacks = self._catalogue_change_callbacks[uuid][f"remove_{name}"]
                        for callback in callbacks:
                            callback(transaction.origin, removed)
                    if added:
                        callbacks = self._catalogue_change_callbacks[uuid][f"add_{name}"]
                        for callback in callbacks:
                            callback(transaction.origin, added)

    def _events_changed(self, events: list[MapEvent], transaction: Transaction) -> None:
        for event in events:
            path = event.path  # type: ignore[attr-defined]
            if len(path) == 0:
                assert isinstance(event, MapEvent)
                keys = event.keys  # type: ignore[attr-defined]
                for uuid in keys:
                    action = keys[uuid]["action"]
                    if action == "delete":
                        for delete_callback in self._event_delete_callbacks[uuid]:
                            delete_callback(transaction.origin)
                        if uuid in self._events:
                            del self._events[uuid]
                        del self._event_delete_callbacks[uuid]
                        self._event_change_callbacks[uuid]
                        del self._event_change_callbacks[uuid]
                    elif action == "add":
                        for create_callback in self._event_create_callbacks:
                            create_callback(transaction.origin, self.get_event(uuid))
            elif len(path) == 1:
                assert isinstance(event, MapEvent)
                uuid = path[0]
                changed_keys = event.keys  # type: ignore[attr-defined]
                for key in changed_keys:
                    if key in self._event_change_callbacks[uuid]:
                        callbacks = self._event_change_callbacks[uuid][key]
                        for callback in callbacks:
                            value = changed_keys[key]["newValue"]
                            model = EventModel.__pydantic_validator__.validate_assignment(EventModel.model_construct(), key, value)
                            callback(transaction.origin, getattr(model, key))
            elif len(path) == 2:
                assert isinstance(event, MapEvent)
                uuid, name = path
                added = {}
                removed = set()
                keys = event.keys  # type: ignore[attr-defined]
                for key, val in keys.items():
                    if val["action"] == "delete":
                        removed.add(key)
                    elif val["action"] == "add":
                        added[key] = val["newValue"]
                    elif val["action"] == "update":
                        added[key] = val["newValue"]
                if removed:
                    callbacks = self._event_change_callbacks[uuid][f"remove_{name}"]
                    for callback in callbacks:
                        callback(transaction.origin, removed)
                if added:
                    callbacks = self._event_change_callbacks[uuid][f"add_{name}"]
                    for callback in callbacks:
                        callback(transaction.origin, added)

    @property
    def catalogues(self) -> set[Catalogue]:
        """
        Returns:
            The catalogues in the database.
        """
        return {Catalogue.from_map(catalogue, self) for catalogue in self._catalogue_maps.values()}

    @property
    def events(self) -> set[Event]:
        """
        Returns:
            The events in the database.
        """
        return {Event.from_map(event, self) for event in self._event_maps.values()}

    def create_catalogue(
        self,
        *,
        name: str,
        author: str,
        uuid: UUID | str | bytes | bytearray | None = None,
        tags: list[str] | None = None,
        attributes: dict[str, Any] | None = None,
        events: Iterable[Event] | Event | None = None,
    ) -> Catalogue:
        """
        Creates a catalogue in the database.

        Args:
            name: The name of the catalogue.
            author: The author of the catalogue.
            uuid: The optional UUID of the catalogue.
            tags: The optional tags of the catalogue.
            attributes: The optional attributes of the catalogue.
            events: The initial event(s) in the catalogue.

        Returns:
            The created [Catalogue][cocat.Catalogue].
        """
        with self.transaction():
            kwargs: dict[str, Any] = {
                "name": name,
                "author": author,
            }
            if uuid is not None:
                kwargs["uuid"] = uuid
            if tags is not None:
                kwargs["tags"] = tags
            if attributes is not None:
                kwargs["attributes"] = attributes
            model = CatalogueModel(**kwargs)
            catalogue = Catalogue.new(model, self)
            self._catalogue_maps[str(model.uuid)] = catalogue._map
            if events is not None:
                if isinstance(events, Event):
                    events = [events]
                catalogue.add_events(events)

        return catalogue

    def create_event(
        self,
        *,
        start: datetime | int | float | str,
        stop: datetime | int | float | str,
        author: str,
        uuid: UUID | str | bytes | bytearray | None = None,
        tags: list[str] | None = None,
        products: list[str] | None = None,
        rating: int | None = None,
        attributes: dict[str, Any] | None = None,
    ) -> Event:
        """
        Creates an event in the database.

        Args:
            start: The start date of the event.
            stop: The stop date of the event.
            author: The author of the event.
            uuid: The optional UUID of the event.
            tags: The optional tags of the event.
            products: The optional products of the event.
            rating: The optional rating of the event.
            attributes: The optional attributes of the catalogue.

        Returns:
            The created [Event][cocat.Event].
        """
        with self.transaction():
            kwargs: dict[str, Any] = {
                "author": author,
                "start": start,
                "stop": stop,
            }
            if uuid is not None:
                kwargs["uuid"] = uuid
            if tags is not None:
                kwargs["tags"] = tags
            if attributes is not None:
                kwargs["attributes"] = attributes
            if products is not None:
                kwargs["products"] = products
            if rating is not None:
                kwargs["rating"] = rating
            model = EventModel(**kwargs)
            event = Event.new(model, self)
            self._event_maps[str(model.uuid)] = event._map
            return event

    def on_create_catalogue(self, callback: Callable[[Catalogue], None]) -> None:
        """
        Registers a callback to be called when a catalogue is created.

        Args:
            callback: The callback to call with the created catalogue.
        """
        self._catalogue_create_callbacks.append(partial(self._callback, callback))

    def on_create_event(self, callback: Callable[[Event], None]) -> None:
        """
        Registers a callback to be called when an event is created.

        Args:
            callback: The callback to call with the created event.
        """
        self._event_create_callbacks.append(partial(self._callback, callback))

    def get_catalogue(self, uuid_or_name: str) -> Catalogue:
        """
        Args:
            uuid_or_name: The UUID of the catalogue to get, or its name.

        Returns:
            The catalogue with the given UUID or name.
        """
        try:
            catalogue = Catalogue.from_uuid(uuid_or_name, self)
        except KeyError:
            for uuid in self._catalogue_maps:
                if self._catalogue_maps[uuid]["name"] == uuid_or_name:
                    catalogue = Catalogue.from_uuid(uuid, self)
                    break
            else:
                raise RuntimeError(f"No catalogue found with name or UUID: {uuid_or_name}")
        return catalogue

    def get_event(self, uuid: str) -> Event:
        """
        Args:
            uuid: The UUID of the event to get.

        Returns:
            The event with the given UUID.
        """
        try:
            return Event.from_uuid(uuid, self)
        except KeyError:
            raise RuntimeError(f"No event found with UUID: {uuid}")

    def _handle_sync_message(self, message: bytes, db: "DB", init: bool = False) -> None:
        if init:
            _message = create_sync_message(self._doc)
            db._handle_sync_message(_message, self)

        message_type = message[0]
        if message_type == YMessageType.SYNC:
            try:
                reply = handle_sync_message(message[1:], self._doc)
                if reply is not None:
                    db._handle_sync_message(reply, self)
            except RuntimeError as exc:
                if str(exc) not in ("Already mutably borrowed", "Already in a transaction"):  # pragma: nocover
                    raise

    def sync(self, db: "DB") -> None:
        """
        Keeps the database in sync with another database. Mostly used for tests.

        Args:
            db: The database to keep in sync with this one.
        """
        if db in self._synced or self in db._synced:
            return

        self._synced.append(db)
        message = create_sync_message(self._doc)
        db._handle_sync_message(message, self, True)

        self._doc.observe(partial(send_update, db, self))
        db._doc.observe(partial(send_update, self, db))

    def to_dict(self) -> dict[str, Any]:
        """
        Returns:
            The database as a dictionary.
        """
        return {
            "catalogues": [catalogue.to_dict() for catalogue in self.catalogues],
            "events": [event.to_dict() for event in self.events],
        }

    def to_json(self) -> str:
        """
        Returns:
            The database as a JSON string.
        """
        return json.dumps(self.to_dict())


def send_update(destination: DB, source: DB, event: TransactionEvent) -> None:
    message = create_update_message(event.update)
    destination._handle_sync_message(message, source)



import sys
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from json import dumps
from typing import Any, TYPE_CHECKING

from pycrdt import Map

from .base import Mixin
from .models import EventModel

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: nocover
    from typing_extensions import Self

if TYPE_CHECKING:
    from .db import DB


@dataclass(eq=False)
class Event(Mixin):
    _uuid: str
    _map: Map
    _db: "DB"

    def _check_deleted(self):
        if self._uuid not in self._db._event_maps:
            raise RuntimeError("Event has been deleted")

    def __eq__(self, other: Any) -> bool:
        self._check_deleted()
        if not isinstance(other, Event):
            return NotImplemented

        return self._uuid == other._uuid

    def __repr__(self) -> str:
        return dumps(self.to_dict())

    def __hash__(self) -> int:
        return hash(self._uuid)

    def _get(self, name: str) -> Any:
        self._check_deleted()
        value = self._map[name]
        model = EventModel.__pydantic_validator__.validate_assignment(EventModel.model_construct(), name, value)
        return getattr(model, name)

    def _set(self, name: str, value: Any, func: Callable[[Any], Any] | None = None) -> None:
        with self._db.transaction():
            self._check_deleted()
            model = EventModel.__pydantic_validator__.validate_assignment(EventModel.model_construct(), name, value)
            val = getattr(model, name)
            if func is not None:
                val = func(val)
            self._map[name] = val

    def _on_change(self, name: str, callback: Callable[[Any], None]) -> None:
        self._check_deleted()
        self._db._event_change_callbacks[self._uuid][name].append(partial(self._callback, callback))

    def _on_add(self, field: str, callback: Callable[[Any], None]) -> None:
        self._check_deleted()
        self._db._event_change_callbacks[self._uuid][f"add_{field}"].append(partial(self._callback, callback))

    def _on_remove(self, field: str, callback: Callable[[list[str]], None]) -> None:
        self._check_deleted()
        self._db._event_change_callbacks[self._uuid][f"remove_{field}"].append(partial(self._callback, callback))

    @classmethod
    def new(cls, model: EventModel, db: "DB") -> Self:
        uuid = str(model.uuid)
        map = Map(dict(
            uuid=uuid,
            start=str(model.start),
            stop=str(model.stop),
            author=model.author,
            tags=Map({val: True for val in model.tags}),
            products=Map({val: True for val in model.products}),
            rating=model.rating,
            attributes=Map(model.attributes),
        ))
        self = cls(uuid, map, db)
        db._events[uuid] = self
        return self

    @classmethod
    def from_map(cls, map: Map, db: "DB") -> Self:
        uuid = map["uuid"]
        self = cls(uuid, map, db)
        db._events[uuid] = self
        return self

    @classmethod
    def from_uuid(cls, uuid: str, db: "DB") -> Self:
        map = db._event_maps[uuid]
        self = cls(uuid, map, db)
        db._events[uuid] = self
        return self

    def to_dict(self) -> dict[str, Any]:
        """
        Returns:
            The event as a dictionary.
        """
        self._check_deleted()
        dct = self._map.to_py()
        assert dct is not None
        dct["tags"] = list(dct["tags"].keys())
        dct["products"] = list(dct["products"].keys())
        dct["attributes"] = dict(sorted(dct["attributes"].items()))
        return dict(sorted(dct.items()))

    def on_change_author(self, callback: Callable[[Any], None]) -> None:
        """
        Registers a callback to be called when the event author changes.

        Args:
            callback: The callback to call with the new author.
        """
        self._on_change("author", callback)

    def on_change_start(self, callback: Callable[[datetime], None]) -> None:
        """
        Registers a callback to be called when the event start date changes.

        Args:
            callback: The callback to call with the new start date.
        """
        self._on_change("start", callback)

    def on_change_stop(self, callback: Callable[[datetime], None]) -> None:
        """
        Registers a callback to be called when the event stop date changes.

        Args:
            callback: The callback to call with the new stop date.
        """
        self._on_change("stop", callback)

    def on_change_range(self, callback: Callable[[datetime, datetime], None]) -> None:
        """
        Registers a callback to be called when the event start and/or stop dates change.

        Args:
            callback: The callback to call with the new range.
        """
        range: list[None | datetime] = [None, None]

        def callback_start(start: datetime) -> None:
            if range != [start, self.stop]:
                range[0], range[1] = start, self.stop
                callback(start, self.stop)

        def callback_stop(stop: datetime) -> None:
            if range != [self.start, stop]:
                range[0], range[1] = self.start, stop
                callback(self.start, stop)

        self._on_change("start", callback_start)
        self._on_change("stop", callback_stop)

    def on_change_rating(self, callback: Callable[[Any], None]) -> None:
        """
        Registers a callback to be called when the event rating changes.

        Args:
            callback: The callback to call with the new rating.
        """
        self._on_change("rating", callback)

    def on_delete(self, callback: Callable[[], None]) -> None:
        """
        Registers a callback to be called when the event is removed from the database.

        Args:
            callback: The callback to call.
        """
        self._check_deleted()
        self._db._event_delete_callbacks[self._uuid].append(partial(self._callback, callback))

    def delete(self):
        """
        Removes the event from the database.
        """
        with self._db.transaction():
            self._check_deleted()
            del self._db._event_maps[self._uuid]
            for uuid, catalogue in self._db._catalogue_maps.items():
                catalogue_events = catalogue["events"]
                if self._uuid in catalogue_events:
                    del catalogue_events[self._uuid]

    @property
    def start(self) -> datetime:
        """
        Returns:
            The start date of the event.
        """
        return self._get("start")

    @start.setter
    def start(self, value: datetime) -> None:
        """
        Args:
            value: The start date of the event to set.
        """
        self._set("start", value, str)

    @property
    def stop(self) -> datetime:
        """
        Returns:
            The stop date of the event.
        """
        return self._get("stop")

    @stop.setter
    def stop(self, value: datetime) -> None:
        """
        Args:
            value: The stop date of the event to set.
        """
        self._set("stop", value, str)

    @property
    def range(self) -> tuple[datetime, datetime]:
        """
        Returns:
            The start and stop dates of the event.
        """
        return self._get("start"), self._get("stop")

    @range.setter
    def range(self, value: tuple[datetime, datetime]) -> None:
        """
        Args:
            value: The start and stop dates of the event to set.
        """
        with self._db.transaction():
            start, stop = value
            self._set("start", start, str)
            self._set("stop", stop, str)

    @property
    def rating(self) -> int:
        """
        Returns:
            The rating of the event.
        """
        return self._get("rating")

    @rating.setter
    def rating(self, value: int) -> None:
        """
        Args:
            value: The rating of the event to set.
        """
        self._set("rating", value)

    @property
    def products(self) -> set[str]:
        """
        Returns:
            The products of the event.
        """
        return set(self._get_from_map("products"))

    @products.setter
    def products(self, value: set[str]) -> None:
        """
        Args:
            value: The products of the event to set.
        """
        products = {val: True for val in value}
        self._set_in_map("products", products)

    def on_add_products(self, callback: Callable[[set[str]], None]) -> None:
        """
        Registers a callback to be called when products are added.

        Args:
            callback: The callback to call with the added products.
        """

        def _callback(values: dict[str, Any]) -> None:
            callback(set(values))

        self._on_add("products", _callback)

    def on_remove_products(self, callback: Callable[[list[str]], None]) -> None:
        """
        Registers a callback to be called when products are removed.

        Args:
            callback: The callback to call with the removed products.
        """
        self._on_remove("products", callback)

    def add_products(self, keys: Iterable[str] | str) -> None:
        """
        Adds product(s) to the event.

        Args:
            keys: The products to add to the event.
        """
        self._add_keys("products", keys)

    def remove_products(self, keys: Iterable[str] | str) -> None:
        """
        Removed product(s) from the event.

        Args:
            keys: The products to remove from the event.
        """
        self._remove_keys("products", keys)



import contextlib

from anycorn import Config, serve as anycorn_serve
from anyio import Event, run
from cyclopts import App
from fastapi_users.exceptions import UserAlreadyExists

from .app.app import CocatApp
from .app.db import create_db_and_tables, get_async_session, get_user_db
from .app.schemas import UserCreate
from .app.users import get_user_manager

get_async_session_context = contextlib.asynccontextmanager(get_async_session)
get_user_db_context = contextlib.asynccontextmanager(get_user_db)
get_user_manager_context = contextlib.asynccontextmanager(get_user_manager)

app = App()

@app.command
def serve(
    *,
    host: str = "localhost",
    port: int = 8000,
    update_dir: str = "",
    db_path: str = "./test.db",
):
    """
    Launch a server.

    Args:
        host: The server host name
        port: The server port number
        update_dir: The path to the directory where the room updates are saved
        db_path: The path to the user database
    """
    run(_serve, host, port, update_dir, db_path)


@app.command
def create_user(
    *,
    email: str,
    password: str,
    is_superuser: bool = False,
    db_path: str = "./test.db",
):
    """
    Create a new user in the database.

    Args:
        email: The user e-mail
        password: The user password
        is_superuser: Whether the user is a superuser
        db_path: The path to the user database
    """
    run(_create_user, email, password, is_superuser, db_path)


async def _serve(host: str, port: int, update_dir: str, db_path: str):
    config = Config()
    config.bind = [f"{host}:{port}"]
    shutdown_event = Event()
    try:
        cocat_app = CocatApp(update_dir, db_path)
        await anycorn_serve(cocat_app.app, config, shutdown_trigger=shutdown_event.wait, mode="asgi")  # type: ignore[arg-type]
    except Exception:
        shutdown_event.set()


async def _create_user(email: str, password: str, is_superuser: bool = False, db_path = "./test.db"):
    await create_db_and_tables(db_path)

    try:
        async with get_async_session_context() as session:
            async with get_user_db_context(session) as user_db:
                async with get_user_manager_context(user_db) as user_manager:
                    user = await user_manager.create(
                        UserCreate(
                            email=email, password=password, is_superuser=is_superuser
                        )
                    )
                    print(f"User created {user}")
                    return user
    except UserAlreadyExists:
        print(f"User {email} already exists")
        raise


def main():
    app()



from collections.abc import Callable, Iterable
from typing import Any, TYPE_CHECKING, cast
from uuid import UUID

from pycrdt import Map

if TYPE_CHECKING:
    from .db import DB

class Mixin:
    _uuid: str
    _map: Map
    _db: "DB"
    _get: Callable[[str], Any]
    _set: Callable[[str, Any], None]
    _check_deleted: Callable[[], None]
    _on_add: Callable[[str, Callable[[Any], None]], None]
    _on_remove: Callable[[str, Callable[[list[str]], None]], None]

    def _callback(self, callback: Callable[..., None], origin: Any, *args: Any) -> None:
        if origin is not self:
            callback(*args)

    def _get_from_map(self, field: str) -> dict[str, Any]:
        self._check_deleted()
        map = cast(Map, self._map[field])
        res = map.to_py()
        assert res is not None
        return res

    def _set_in_map(self, field: str, value: dict[str, Any]) -> None:
        with self._db.transaction():
            self._check_deleted()
            map = cast(Map, self._map[field])
            map.clear()
            map.update(value)

    def _add_keys(self, field: str, keys: Iterable[str] | str) -> None:
        with self._db.transaction():
            key_list = [keys] if isinstance(keys, str) else keys
            self._check_deleted()
            map = cast(Map, self._map[field])
            for key in key_list:
                map[key] = True

    def _add_items(self, field: str, items: dict[str, Any]) -> None:
        with self._db.transaction():
            self._check_deleted()
            map = cast(Map, self._map[field])
            map.update(items)

    def _remove_keys(self, field: str, keys: Iterable[str] | str) -> None:
        with self._db.transaction():
            key_list = [keys] if isinstance(keys, str) else keys
            self._check_deleted()
            map = cast(Map, self._map[field])
            for key in key_list:
                del map[key]

    def on_set_attributes(self, callback: Callable[[dict[str, Any]], None]) -> None:
        """
        Registers a callback to be called when attributes are set.

        Args:
            callback: The callback to call with a dictionary of attribute items that were set.
        """
        self._on_add("attributes", callback)

    def on_remove_attributes(self, callback: Callable[[list[str]], None]) -> None:
        """
        Registers a callback to be called when attributes are removed.

        Args:
            callback: The callback to call with a list of attribute keys that were removed.
        """
        self._on_remove("attributes", callback)

    def set_attributes(self, **kwargs: Any) -> None:
        """
        Sets attributes.

        Args:
            kwargs: The attributes to set.
        """
        self._add_items("attributes", kwargs)

    def remove_attributes(self, keys: Iterable[str] | str) -> None:
        """
        Removes attributes.

        Args:
            keys: The attribute keys to remove.
        """
        self._remove_keys("attributes", keys)

    def on_add_tags(self, callback: Callable[[set[str]], None]) -> None:
        """
        Registers a callback to be called when tags are added.

        Args:
            callback: The callback to call with the set of added tags.
        """

        def _callback(values: dict[str, Any]) -> None:
            callback(set(values))

        self._on_add("tags", _callback)

    def on_remove_tags(self, callback: Callable[[list[str]], None]) -> None:
        """
        Registers a callback to be called when tags are removed:

        Args:
            callback: The callback to call with the list of removed tags.
        """
        self._on_remove("tags", callback)

    def add_tags(self, keys: Iterable[str] | str) -> None:
        """
        Adds tags.

        Args:
            keys: The tags to add.
        """
        self._add_keys("tags", keys)

    def remove_tags(self, keys: Iterable[str] | str) -> None:
        """
        Removes tags.

        Args:
            keys: The tags to remove.
        """
        self._remove_keys("tags", keys)

    @property
    def db(self) -> "DB":
        """
        Returns:
            The [database][cocat.DB] it belongs to.
        """
        return self._db

    @property
    def uuid(self) -> UUID:
        """
        Returns:
            The UUID.
        """
        return UUID(self._uuid)

    @property
    def tags(self) -> set[str]:
        """
        Returns:
            The tags.
        """
        return set(self._get_from_map("tags"))

    @tags.setter
    def tags(self, value: set[str]) -> None:
        """
        Args:
            value: The tags to set.
        """
        tags = {val: True for val in value}
        self._set_in_map("tags", tags)

    @property
    def attributes(self) -> dict[str, Any]:
        """
        Returns:
            The attributes.
        """
        return self._get_from_map("attributes")

    @attributes.setter
    def attributes(self, value: dict[str, Any]) -> None:
        """
        Args:
            value: The attributes to set.
        """
        self._set_in_map("attributes", value)

    @property
    def author(self) -> str:
        """
        Returns:
            The author.
        """
        return self._get("author")

    @author.setter
    def author(self, value: str) -> None:
        """
        Args:
            value: The author to set.
        """
        self._set("author", value)
